
#include <sys/types.h>
#include <sys/param.h>
#include <sys/proc.h>
#include <sys/module.h>
#include <sys/sysent.h>
#include <sys/kernel.h>
#include <sys/systm.h>

#include <sys/vmmeter.h>
#include <sys/malloc.h>
#include <sys/proc.h>

#include <mydrv.h>

struct pcb* mycurpcb;
struct vmmeter vmm;
void** kmembuf = 0;

static int
mydrv_entry (struct proc* p, void** uap)
{
  int s;
  int loop;
  int error=0;
  mydrv_args_t arg;

  copyin (*uap, &arg, sizeof(arg)) ;

  printf("mydrv:entry: cmd 0x%x size %d loop %d\n", arg.cmd, arg.size, arg.loop);

  switch (arg.cmd)
  {
    case MYCMD_MALLOC:
    {
      printf("mydrv:malloc: kmembuf %p curpcb %p [%d]\n", kmembuf, curpcb, __LINE__);

      if (kmembuf)
      {
	error= EINVAL;
	break;
      }

      s = splnet();
      printf("mydrv:malloc: kmembuf %p curpcb %p [%d]\n", kmembuf, curpcb, __LINE__);
      mycurpcb = curpcb;
      vmm = cnt;
      MALLOC(kmembuf,void**,arg.loop*sizeof(void*),M_DEVBUF,M_NOWAIT);
      printf("mydrv:malloc: kmembuf %p curpcb %p [%d]\n", kmembuf, curpcb, __LINE__);

      if (kmembuf==0)
      {
	printf("mydrv:malloc: FAILURE kmembuf: free %d wire %d active %d inactive %d cache %d\n",
	  vmm.v_free_count, vmm.v_wire_count, vmm.v_active_count, vmm.v_inactive_count, vmm.v_cache_count);
	splx(s);
	error= ENOMEM;
	break;
      }

      for (loop=0; loop < arg.loop; loop++)
      {
	vmm = cnt;
	MALLOC(kmembuf[loop],void*,arg.size,M_DEVBUF,M_NOWAIT);

	if (kmembuf[loop]==0)
	{
	  printf("mydrv:malloc: FAILURE kmembuf[%d]: free %d wire %d active %d inactive %d cache %d\n", loop,
	    vmm.v_free_count, vmm.v_wire_count, vmm.v_active_count, vmm.v_inactive_count, vmm.v_cache_count);
	  splx(s);
	  error= ENOMEM;
	  break;
	}
      }

      mycurpcb = 0;
      splx(s);

      printf("mydrv:malloc: SUCCESS kmembuf %p: free %d wire %d active %d inactive %d cache %d\n", kmembuf,
	vmm.v_free_count, vmm.v_wire_count, vmm.v_active_count, vmm.v_inactive_count, vmm.v_cache_count);

      break;
    }
    case MYCMD_FREE:
    {
      printf("mydrv:free: kmembuf %p\n", kmembuf);

      if (!kmembuf)
      {
	error= EINVAL;
	break;
      }

      for (loop=0; loop < arg.loop; loop++)
      {
	if (kmembuf[loop]) FREE((caddr_t)kmembuf[loop],M_DEVBUF);
      }

      FREE((caddr_t)kmembuf, M_DEVBUF);

      printf("mydrv:free: SUCCESS kmembuf %p: free %d wire %d active %d inactive %d cache %d\n", kmembuf,
	vmm.v_free_count, vmm.v_wire_count, vmm.v_active_count, vmm.v_inactive_count, vmm.v_cache_count);

      kmembuf = 0 ;

      break;
    }
    default:
    {
      error=EINVAL;
      break;
    }
  }

  return error;
}

static struct sysent mydrv_sysent =
{
  1,
  (int(*)(struct proc*, void*)) mydrv_entry
} ;

static int offset = NO_SYSCALL;

static int
mydrv_load (struct module* mod, int cmd, void* args)
{
  int error=0;

  switch(cmd)
  {
    case MOD_LOAD:
    {
      printf("mydrv:load: LOAD called: offset %d\n", offset);
      break;
    }
    case MOD_UNLOAD:
    {
      printf("mydrv:load: UNLOAD called: offset %d\n", offset);
      break;
    }
    default:
    {
      error=EINVAL;
      break;
    }
  }

  return error;
}

SYSCALL_MODULE(mydrv,&offset,&mydrv_sysent,mydrv_load,NULL);

